/*
  Pins.h Directly access GPIO ports efficiently
  Copyright (C) 2020  Thomas Almy

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

  Contact the author at tom@almy.us */
#ifndef _PINS_H
#define _PINS_H
#include <avr/io.h>

#if defined (__AVR_ATmega4809__)
// Arduino Nano Every

/* The gcc compiler will compress reads and stores of these registers
 * to single (atomic) instructions because addresses are in range to do
 * so. It is not obvious that something like VPORTA.DIR |= 1 << 3 will do
 * the same instead of a load, or, store sequence, however it does. 
 */

struct bits
{
  /* VPORTA.DIR at 0x00, VPORTA.OUT at 0x01, VPORTA.IN at 0x02, VPORTA.INTFLAGS at 0x03 */
  int digital_2:1;
  int digital_7:1;
  int analog_4:1;
  int analog_5:1;
  int :1;
  int :1;
  int :1;  
  int :1;
  
  /* Skip three bytes to get to Port B's register */
  int :16;
  int :8;
  
  /* VPORTB.DIR at 0x04, VPORTB.OUT at 0x05, VPORTB.IN at 0x06, VPORTB.INTFLAGS at 0x07 */
  int digital_9:1;
  int digital_10:1;
  int digital_5:1;
  int :1;
  int :1;
  int :1;
  int :1;  
  int :1;
  
  /* Skip three bytes to get to Port C's register */
  int :16;
  int :8;

  /* VPORTC as above */
  int :1;
  int :1;
  int :1;
  int :1;
  int digital_0:1;
  int digital_1:1;
  int digital_4:1;  
  int :1;
  
  /* Skip three bytes to get to Port D's register */
  int :16;
  int :8;

  /* VPORTD as above */
  int analog_3:1;
  int analog_2:1;
  int analog_1:1;
  int analog_0:1;
  int analog_6:1;
  int analog_7:1;
  int :1;  
  int aref_pin:1;
  
  /* Skip three bytes to get to Port E's register */
  int :16;
  int :8;
  
  /* VPORTE as above */
  int digital_11:1;
  int digital_12:1;
  int digital_13:1;
  int digital_8:1;
  int :1;
  int :1;
  int :1;  
  int :1;
  
  /* Skip three bytes to get to Port F's register */
  int :16;
  int :8;
  
    /* VPORTF as above */
  int :1;
  int :1;
  int :1;
  int :1;
  int digital_6:1;
  int digital_3:1;
  int reset_pin:1;  
  int :1;
  };


/* These macros are used to access the different register bits.
   For example port.digital_8  gives VPORTB.OUT bit 0
   I'm using the same names for these as the other AVRs. */

#define ddr (*(/*volatile*/ struct bits *)(&VPORTA.DIR))
#define pin (*(volatile struct bits *)(&VPORTA.IN))
#define port (*(volatile struct bits *)(&VPORTA.OUT))

#define intflg (*(volatile struct bits *)(&VPORTA.INTFLAGS))


/* not defining ddr for volatile means that setting consecutive bits
 * can be done more efficiently, however in that case it won't be
 * atomic. This will never be a problem unless ddr is changed after
 * enabling interrupts. */

#else
#ifdef DDRE // Arduino Mega
/* The gcc compiler will compress reads and stores of these registers
 * to single (atomic) instructions because addresses are in range to do
 * so. It is not obvious that something like PORTB |= 1 << 3 will do
 * the same instead of a load, or, store sequence, however it does. 
 * In processors like ATmega2560 (Arduino Mega 2560)
 * not all the ports are in this good address range so operations will
 * only be atomic for ports A through G */

struct bits
{
  /* PINA at 0x00, DDRA at 0x01, PORTA at 0x02 */
  int digital_22:1;
  int digital_23:1;
  int digital_24:1;
  int digital_25:1;
  int digital_26:1;
  int digital_27:1;
  int digital_28:1;
  int digital_29:1; 
  
  /* Skip two bytes to get to Port B's register */
  int :16;
  
  /* PINB at 0x03, DDRB at 0x04, PORTB at 0x05 */
  int digital_53:1;
  int digital_52:1;
  int digital_51:1;
  int digital_50:1;
  int digital_10:1;
  int digital_11:1;
  int digital_12:1;
  int digital_13:1;
  
  /* Skip two bytes to get to Port C's register */
  int :16;

  /* PINC at 0x06, DDRC at 0x07, PORTC at 0x08 */
  int digital_37:1;
  int digital_36:1;
  int digital_35:1;
  int digital_34:1;
  int digital_33:1;
  int digital_32:1;
  int digital_31:1;
  int digital_30:1;

  /* Skip two bytes to get to port D's register. */
  int :16;

  /* PIND at 0x09, DDRD at 0x0a, PORTD at 0x0b */
  int digital_21:1;
  int digital_20:1;
  int digital_19:1;
  int digital_18:1;
  int :3; // Not accessible on Arduino board
  int digital_38:1;

  /* Skip two bytes to get to Port E's register */
  int :16;

  /* PINE at 0x0c, DDRE at 0x0d, PORTE at 0x0e */
  int digital_0:1;
  int digital_1:1;
  int :1; // Not accessible on Arduino board
  int digital_5:1;
  int digital_2:1;
  int digital_3:1;
  int :2; // Not accessible on Arduino board

  /* Skip two bytes to get to Port F's register */
  int :16;

  /* PINF at 0x0f, DDRF at 0x10, PORTF at 0x11 */
  int analog_0:1;
  int analog_1:1;
  int analog_2:1;
  int analog_3:1;
  int analog_4:1;
  int analog_5:1;
  int analog_6:1;
  int analog_7:1;

  /* Skip two bytes to get to Port G's register */
  int :16;

  /* PING at 0x12, DDRG at 0x13, PORTG at 0x14 */
  int digital_41:1;
  int digital_40:1;
  int digital_39:1;
  int :2; // Not accessible on Arduino board
  int digital_4:1;
  int :2; // Not used
  
  /* Skip 205 bytes to get to port H's registers. */
  int :8;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  int :16;
  
  /* PINH at 0x100, DDRH at 0x101, PORTH at 0x102 */
  int digital_17:1;
  int digital_16:1;
  int :1; // Not accessible on Arduino board
  int digital_6:1;
  int digital_7:1;
  int digital_8:1;
  int digital_9:1;
  int :1; // Not accessible on Arduino board

  /* Skip two bytes to get to port J's register. */
  int :16;
  /* PINJ at 0x103, DDRJ at 0x104, PORTJ at 0x105 */
  int digital_15:1;
  int digital_14:1;
  int :6; // Not accessible on Arduino board

  /* Skip two bytes to get to port K's register. */
  int :16;

  /* PINK at 0x106, DDRK at 0x107, PORTK at 0x108 */
  int analog_8:1;
  int analog_9:1;
  int analog_10:1;
  int analog_11:1;
  int analog_12:1;
  int analog_13:1;
  int analog_14:1;
  int analog_15:1;

  /* Skip two bytes to get to port L's register. */
  int :16;

  /* PINL at 0x109, DDRL at 0x10a, PORTL at 0x10b */
  int digital_49:1;
  int digital_48:1;
  int digital_47:1;
  int digital_46:1;
  int digital_45:1;
  int digital_44:1;
  int digital_43:1;
  int digital_42:1;

};


/* These macros are used to access the different register bits.
   For example port.digital_8  gives PORTB bit 0*/

#define pin (*(volatile struct bits *)(&PINA))
#define ddr (*(/*volatile*/ struct bits *)(&DDRA))
#define port (*(volatile struct bits *)(&PORTA))

/* not defining ddr for volatile means that setting consecutive bits
 * can be done more efficiently, however in that case it won't be
 * atomic. This will never be a problem unless ddr is changed after
 * enabling interrupts. */

#else
// Arduino UNO

/* The gcc compiler will compress reads and stores of these registers
 * to single (atomic) instructions because addresses are in range to do
 * so. It is not obvious that something like PORTB |= 1 << 3 will do
 * the same instead of a load, or, store sequence, however it does. 
 * In processors like ATmega2560 (Arduino Mega 2560)
 * not all the ports are in this good address range so operations will
 * only be atomic for ports A through G */

struct bits
{
  /* PINB at 0x03, DDRB at 0x04, PORTB at 0x05 */
  int digital_8:1;
  int digital_9:1;
  int digital_10:1;
  int digital_11:1;
  int digital_12:1;
  int digital_13:1;
  int :2;  /* PB6 and PB7 are not accessible and reserved for the crystal */
  
  /* Skip two bytes to get to Port C's register */
  int :16;

  /* PINC at 0x06, DDRC at 0x07, PORTC at 0x08 */
  int analog_0:1;
  int analog_1:1;
  int analog_2:1;
  int analog_3:1;
  int analog_4:1;
  int analog_5:1;
  int :2;  /* PC6 is not accessible (crystal) and PC7 is not used */

  /* Skip two bytes to get to port D's register. */
  int :16;

  /* PIND at 0x09, DDRD at 0x0a, PORTD at 0x0b */
  int digital_0:1;
  int digital_1:1;
  int digital_2:1;
  int digital_3:1;
  int digital_4:1;
  int digital_5:1;
  int digital_6:1;
  int digital_7:1;

};


/* These macros are used to access the different register bits.
   For example port.digital_8  gives PORTB bit 0*/

#define pin (*(volatile struct bits *)(&PINB))
#define ddr (*(/*volatile*/ struct bits *)(&DDRB))
#define port (*(volatile struct bits *)(&PORTB))

/* not defining ddr for volatile means that setting consecutive bits
 * can be done more efficiently, however in that case it won't be
 * atomic. This will never be a problem unless ddr is changed after
 * enabling interrupts. */

#endif
#endif
#endif // _PINS_H
