/*
  SPI_SLAVE -- Demonstrates use of SPI slave mode. Two bytes in serial
  chain, interrupt driven. To demonstrate, implements a master SPI interface
  in software and connects them together.

  Copyright (C) 2020  Thomas Almy

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

  Contact the author at tom@almy.us */
// The corresponding master and slave signals must be wired together on the board
#define MASTER_TOO
#include <Pins.h>
#define SLAVE_MOSI digital_11
#define SLAVE_MISO digital_12
#define SLAVE_SS digital_8 // Is PE3
#define SLAVE_SCK digital_13
#ifdef MASTER_TOO
#define MASTER_MOSI digital_2
#define MASTER_MISO digital_3
#define MASTER_SS digital_4
#define MASTER_SCK digital_5
#endif

void setup() {
#ifdef MASTER_TOO
  Serial.begin(9600);
  ddr.MASTER_MOSI = 1;
  ddr.MASTER_SS = 1;
  port.MASTER_SS = 1;
  ddr.MASTER_SCK = 1;
#endif
  PORTMUX.TWISPIROUTEA |= PORTMUX_SPI0_ALT2_gc;
  ddr.SLAVE_MISO = 1; // Config output pins
  PORTE.PIN3CTRL = PORT_ISC_RISING_gc; // Slave SS pin interrupts on rising edge
  SPI0.CTRLA = 0x41; // Enable in slave mode, LSB first
  SPI0.CTRLB = 0xc0; // Enable buffering
  SPI0.INTCTRL = 0x80; // Interrupt when receive buffer has content
  SPI0.DATA = 0; // First two bytes to send.
  SPI0.DATA = 0;
}

#ifdef MASTER_TOO
uint8_t masterOutData[4] = {0x55, 0xaa, 0x21, 0x31};
uint8_t masterInData[4] = {0, 0, 0, 0};
uint8_t masterIndex;

uint8_t transferData(uint8_t val) {
  // This function combines shiftIn and shiftOut
  uint8_t retVal = 0;
  uint8_t i;
  for (i = 0; i < 8; i++) {
    port.MASTER_MOSI = (val & 1) != 0; // set bit to send
    val >>= 1;
    port.MASTER_SCK = 1; // clock goes high
    retVal >>= 1;
    retVal |= (pin.MASTER_MISO ? 0x80 : 0); // get bit to read
    port.MASTER_SCK = 0; // clock goes low
  }
  return retVal;
}

void loop() {
  uint8_t i;
  port.MASTER_SS = 0;
  for (i = 0; i < 4; i++) {
    masterInData[i] = transferData(masterOutData[i]);
  }
  port.MASTER_SS = 1;
  Serial.print("Received:");
  for ( i = 0; i < 4; i++) {
    Serial.print(" ");
    Serial.print(masterInData[i], HEX);
  }
  Serial.println("");
  delay(1000);
}
#else
void loop() {} // nothing to do here
#endif

volatile uint8_t inm1, inm2; // buffer last two bytes here
ISR(SPI0_INT_vect) { // A byte (or more) has come in
  // Invoked when data has been transferred
  // Send the previous byte since we want to appear like we have two bytes
  // Note that if this is the last byte to be transferred, we will
  // override what we return in SPI0.DATA in the PORTE ISR. And we will
  // replace the value in inm1 with the seond return value.
  do {
    // Keep last two bytes in
    inm2 = inm1;
    inm1 = SPI0.DATA;
    // data that goes out is that which just came in, because of the hardware
    // transmit buffer already has a byte.
    SPI0.DATA = inm1;
  } while ((SPI0.INTFLAGS & 0x80) != 0); // Repeat until flag clears
}

ISR(PORTE_PORT_vect) { // Interrupt vector for SS rising
  if (intflg.SLAVE_SS != 0) { // SS rising
    while ((SPI0.INTFLAGS & 0x80) != 0) {
      // there is a pending SPI interrupt, so do what it would have done
      //     SPI0.DATA = inm1;
      inm2 = inm1;
      inm1 = SPI0.DATA;
    }
    // Values here are correct!
    SPI0.CTRLA &= ~1; // Turn off SPI to reset
    // Our values are in inm2 (first received) and inm1 (most recent received)
    // Let's return the product of the two bytes.
    uint16_t product = inm2 * inm1;
    inm1 = (uint8_t)(product & 0xff);
    inm2 = (uint8_t)(product >> 8);
    SPI0.CTRLA |= 1; // Turn SPI back on
    // We send the most significant byte first, then the least significant
    SPI0.DATA =  inm2;
    SPI0.DATA = inm1;
    intflg.SLAVE_SS = 1; // reset pin interrupt flag
  }
}
