/*  SPI_SLAVE_BUFFERED -- Demonstrates use of SPI slave mode. Four bytes in serial
  chain, interrupt driven. To demonstrate, implements a master SPI interface
  in software and connects them together.

  Copyright (C) 2020  Thomas Almy

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

  Contact the author at tom@almy.us */
#define MASTER_TOO // Define if this board is also the SPI master

// The corresponding master and slave signals must be wired together on the board
#include <Pins.h>
#define SLAVE_MOSI digital_11
#define SLAVE_MISO digital_12
#define SLAVE_SS digital_8 // Is PE3
#define SLAVE_SCK digital_13
#ifdef MASTER_TOO
#define MASTER_MOSI digital_2
#define MASTER_MISO digital_3
#define MASTER_SS digital_4
#define MASTER_SCK digital_5
#endif

volatile uint8_t receiveBuffer[4];
volatile uint8_t sendBuffer[4];
volatile uint8_t receiveIndex, sendIndex; // Index into above of next byte

void setup() {
#ifdef MASTER_TOO
  Serial.begin(9600);
  ddr.MASTER_MOSI = 1;
  ddr.MASTER_SS = 1;
  port.MASTER_SS = 1;
  ddr.MASTER_SCK = 1;
#endif
  PORTMUX.TWISPIROUTEA |= PORTMUX_SPI0_ALT2_gc;
  ddr.SLAVE_MISO = 1; // Config output pins
  PORTE.PIN3CTRL = PORT_ISC_RISING_gc; // Slave SS pin interrupts on rising edge
  SPI0.CTRLA = 0x41; // Enable in slave mode, LSB first
  SPI0.CTRLB = 0xc0; // Enable buffering
  SPI0.INTCTRL = 0xA0; // Interrupt when receive buffer has content or send buffer empty
}

#ifdef MASTER_TOO
uint8_t masterOutData[4] = {0x55, 0xaa, 0x21, 0x31};
uint8_t masterInData[4] = {0, 0, 0, 0};

uint8_t transferData(uint8_t val) {
  // This function combines shiftIn and shiftOut
  uint8_t retVal = 0;
  uint8_t i;
  for (i = 0; i < 8; i++) {
    port.MASTER_MOSI = (val & 1) != 0; // set bit to send
    val >>= 1;
    port.MASTER_SCK = 1; // clock goes high
    retVal >>= 1;
    retVal |= (pin.MASTER_MISO ? 0x80 : 0); // get bit to read
    port.MASTER_SCK = 0; // clock goes low
  }
  return retVal;
}

void loop() {
  uint8_t i;
  port.MASTER_SS = 0;
  for (i = 0; i < 4; i++) {
    masterInData[i] = transferData(masterOutData[i]);
  }
  port.MASTER_SS = 1;
  Serial.print("Received:");
  for ( i = 0; i < 4; i++) {
    Serial.print(" ");
    Serial.print(masterInData[i], HEX);
  }
  Serial.println("");
  delay(500);
}
#else
void loop() {} // Do nothing in loop if not the master.
#endif

volatile uint8_t dummy;

ISR(SPI0_INT_vect) { // We might be able to send or receive
  if ((SPI0.INTFLAGS & 0x20) != 0) { // transmit data register empty
    if (sendIndex < 4) { // more to send
      SPI0.DATA = sendBuffer[sendIndex];
      sendIndex++;
    } else { // All sent, disable interrupt
      SPI0.INTCTRL &= ~0x20;
    }
  }
  while ((SPI0.INTFLAGS & 0x80) != 0) { // data has arrived
    if (receiveIndex < 4) { // we have a place for it
      receiveBuffer[receiveIndex] = SPI0.DATA;
      receiveIndex++;
    } else { // Why is this here? just throw it out
      dummy = SPI0.DATA;
    }
  }
}

ISR(PORTE_PORT_vect) { // Interrupt vector for SS rising
  if (intflg.SLAVE_SS != 0) { // SS rising
    while ((SPI0.INTFLAGS & 0x80) != 0) {
      // there is a pending SPI interrupt, so do what it would have done
      // It can only be data received at this point
      if (receiveIndex < 4) { // we have a place for it
        receiveBuffer[receiveIndex] = SPI0.DATA;
        receiveIndex++;
      } else { // Why is this here? just throw it out
        dummy = SPI0.DATA;
      }
    }
    // Values here are correct!
    // Our values are in receiveBuffer. Let's "process" them
    for (uint8_t i = 0; i < 4; i++) {
      sendBuffer[i] = receiveBuffer[i] + 1;
    }
    // Reset indices and enable data register empty interrupt
    receiveIndex = 0;
    sendIndex = 0;
    SPI0.CTRLA &= ~1; // turn off
    SPI0.INTCTRL |= 0x20;
    SPI0.CTRLA |= 1; // turn back on

    intflg.SLAVE_SS = 1; // reset pin interrupt flag
  }
}
