/*
  USART_AS_MSPI -- USART0 running in MSPI mode is used to drive the SPI
  module in slave mode. This combines the buffered slave SPI example code
  with new code running a USART as an SPI master.
  Copyright (C) 2020  Thomas Almy

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

  Contact the author at tom@almy.us */
// The corresponding master and slave signals must be wired together on the board
#include <Pins.h>
#define SLAVE_MOSI digital_11
#define SLAVE_MISO digital_12
#define SLAVE_SS digital_8 // Is PE3
#define SLAVE_SCK digital_13

#define MASTER_MOSI digital_2
#define MASTER_MISO digital_7
#define MASTER_SS digital_3
#define MASTER_SCK analog_4

#define BAUDRATE 100000 // 100kHz // The slave SPI limits the maximum speed
#define TWOISRVERSION  // Define if use both RXC and TXC, otherwise use just TXC
// which gives a small performance loss
volatile uint8_t receiveBuffer[4];
volatile uint8_t sendBuffer[4];
volatile uint8_t receiveIndex, sendIndex; // Index into above of next byte

void setup() {
  Serial.begin(9600);
  cli();
  // Configure USART0 as MSPI
  ddr.MASTER_MOSI = 1;
  ddr.MASTER_SS = 1;
  port.MASTER_SS = 1;
  ddr.MASTER_SCK = 1;
  USART0.CTRLC = USART_CMODE_MSPI_gc | 4; // MSPI mode, LSB first
  USART0.BAUD = (32 * F_CPU) / BAUDRATE;
  USART0.CTRLB = 0xc0; // RXEN and TXEN
#ifdef TWOISRVERSION
  USART0.CTRLA = 0xC0; // RXCIE, TXCIE
#else
  USART0.CTRLA = 0x40; // TXCIE
#endif

  // Configure SPI
  PORTMUX.TWISPIROUTEA |= PORTMUX_SPI0_ALT2_gc;
  ddr.SLAVE_MISO = 1; // Config output pins
  PORTE.PIN3CTRL = PORT_ISC_RISING_gc; // Slave SS pin interrupts on rising edge
  SPI0_CTRLA = 0x41; // Enable in slave mode, LSB first
  SPI0_CTRLB = 0xc0; // Enable buffering
  SPI0_INTCTRL = 0xA0; // Interrupt when receive buffer has content or send buffer empty
  sei();
}

uint8_t masterOutData[4] = {0x55, 0xaa, 0x21, 0x31};
uint8_t masterInData[4] = {0, 0, 0, 0};

volatile uint8_t masterIndex;

void loop() {
  uint8_t i;
  port.MASTER_SS = 0;
  masterIndex = 0;
  USART0.TXDATAL = masterOutData[0]; // Start things running
  while (masterIndex < 4) ; // wait until it completes

  Serial.print("Received:");
  for ( i = 0; i < 4; i++) {
    Serial.print(" ");
    Serial.print(masterInData[i], HEX);
  }
  Serial.println("");
  delay(500);
}

#ifdef TWOISRVERSION
ISR(USART0_RXC_vect) { // A byte has come in (and one has gone out)
  masterInData[masterIndex++] = USART0.RXDATAL;
  if (masterIndex < 4) { // send another byte out
    USART0.TXDATAL = masterOutData[masterIndex];
  }
}

ISR(USART0_TXC_vect) { // use this to turn off SS
  port.MASTER_SS = 1;
  USART0.STATUS |= 0x40; // clear the flag
}
#else
ISR(USART0_TXC_vect) { // do it all in one ISR
  masterInData[masterIndex++] = USART0.RXDATAL;
  if (masterIndex < 4) { // send another byte out
    USART0.TXDATAL = masterOutData[masterIndex];
  } else {
    port.MASTER_SS = 1;
  }
  USART0.STATUS |= 0x40; // clear the flag
}
#endif
volatile uint8_t dummy;

ISR(SPI0_INT_vect) { // We might be able to send or receive
  if ((SPI0_INTFLAGS & 0x20) != 0) { // transmit data register empty
    if (sendIndex < 4) { // more to send
      SPI0_DATA = sendBuffer[sendIndex];
      sendIndex++;
    } else { // All sent, disable interrupt
      SPI0_INTCTRL &= ~0x20;
    }
  }
  while ((SPI0_INTFLAGS & 0x80) != 0) { // data has arrived
    if (receiveIndex < 4) { // we have a place for it
      receiveBuffer[receiveIndex] = SPI0_DATA;
      receiveIndex++;
    } else { // Why is this here? just throw it out
      dummy = SPI0_DATA;
    }
  }
}

ISR(PORTE_PORT_vect) { // Interrupt vector for SS rising
  if (intflg.SLAVE_SS != 0) { // SS rising
    while ((SPI0_INTFLAGS & 0x80) != 0) {
      // there is a pending SPI interrupt, so do what it would have done
      // It can only be data received at this point
      if (receiveIndex < 4) { // we have a place for it
        receiveBuffer[receiveIndex] = SPI0_DATA;
        receiveIndex++;
      } else { // Why is this here? just throw it out
        dummy = SPI0_DATA;
      }
    }
    // Values here are correct!
    // Our values are in receiveBuffer. Let's "process" them
    for (uint8_t i = 0; i < 4; i++) {
      sendBuffer[i] = receiveBuffer[i] + 1;
    }
    // Reset indices and enable data register empty interrupt
    receiveIndex = 0;
    sendIndex = 0;
    SPI0_CTRLA &= ~1; // turn off
    SPI0_INTCTRL |= 0x20;
    SPI0_CTRLA |= 1; // turn back on

    intflg.SLAVE_SS = 1; // reset pin interrupt flag
  }
}
