/* zs042_twi -- read and write the I2C devices DS3231 and 24C32 using twi.c
  Copyright (C) 2020  Thomas Almy

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

  Contact the author at tom@almy.us */
extern "C" {
#include "twi.h"
}

#define SETUP // Comment out when not initializing the clock and EEPROM

#define DS3231_ADDRESS 0x68
#define EE_ADDRESS 0x57

void setup()
{
  TWI_MasterInit(100000L); // 100kHz
#ifdef SETUP
  // set the initial time here:
  // DS3231 seconds, minutes, hours, day, date, month, year
  setDS3231time(30, 12, 8, 3, 28, 6, 20);
  // Set the EEPROM contents
  setEEPROM();
#endif
  Serial.begin(9600); // Something requires setting up Serial after the Setup
  // Otherwise the serial port doesn't work.
  // I should figure this out.
  delay(800);
}

void loop()
{
  printEEPROM(); // Display string from EEPROM
  displayTime(); // display the real-time clock data on the Serial Monitor
  delay(1000);
}

/* Write test data to the 24C32 EEPROM */
void setEEPROM(void) {
  uint8_t buff[16];
  buff[0] = 0; // high address
  buff[1] = 0x60; // low address
  strcpy((char *)buff + 2, "Hello World!");
  // number of bytes is 2 plus the length of the string plus 1 for the string delimiter
  TWI_MasterWrite(EE_ADDRESS, buff, strlen((char *)buff + 2) + 3, true); // blocking and send stop at end
}

/* Read and print the test data from the 24C32 EEPROM */
void printEEPROM(void) {
  uint8_t buff[2];
  buff[0] = 0; // high address
  buff[1] = 0x60; // low address
  if (TWI_MasterWrite(EE_ADDRESS, buff, 2, false) // don't send stop at end
      != 0) { // EEPROM not ready after previous write
    return;
  }
  while (true) {
    TWI_MasterRead(EE_ADDRESS, buff, 1, false); // Must read a byte at a time, break out at end of string
    if (buff[0] == 0) break;
    Serial.write(buff[0]);
  }
  TWI_MasterRead(EE_ADDRESS, buff, 0, true); // End transmission with STOP
  Serial.print(" ");
}

/* Set the current time */
void setDS3231time(uint8_t second,
                   uint8_t minute,
                   uint8_t hour,
                   uint8_t dayOfWeek,
                   uint8_t dayOfMonth,
                   uint8_t month,
                   uint8_t year)
{
  uint8_t data[8];
  // sets time and date data to DS3231
  data[0] = 0; // set next input to start at the seconds register
  data[1] = toBCD(second); // set seconds
  data[2] = toBCD(minute); // set minutes
  data[3] = toBCD(hour); // set hours
  data[4] = toBCD(dayOfWeek); // set day of week (1=Sunday, 7=Saturday)
  data[5] = toBCD(dayOfMonth); // set date (1 to 31)
  data[6] = toBCD(month); // set month
  data[7] = toBCD(year); // set year (0 to 99)
  TWI_MasterWrite(DS3231_ADDRESS, data, 8, true); // blocking and send stop at end
}

/* Read the current time */
void readDS3231time(uint8_t *second,
                    uint8_t *minute,
                    uint8_t *hour,
                    uint8_t *dayOfWeek,
                    uint8_t *dayOfMonth,
                    uint8_t *month,
                    uint8_t *year)
{
  uint8_t data[8];
  data[0] = 0;
  TWI_MasterWriteRead(DS3231_ADDRESS, data, 1, 7, true);
  // request seven bytes of data from DS3231 starting from register 00h
  *second = fromBCD(data[0] & 0x7f);
  *minute = fromBCD(data[1]);
  *hour = fromBCD(data[2] & 0x3f);
  *dayOfWeek = fromBCD(data[3]);
  *dayOfMonth = fromBCD(data[4]);
  *month = fromBCD(data[5]);
  *year = fromBCD(data[6]);
}

/* Read and display the current time */
void displayTime()
{
  uint8_t second, minute, hour, dayOfWeek, dayOfMonth, month, year;
  // retrieve data from DS3231
  readDS3231time(&second, &minute, &hour, &dayOfWeek, &dayOfMonth, &month,
                 &year);
  // send it to the serial monitor
  Serial.print(hour);
  // convert the uint8_t variable to a decimal number when displayed
  Serial.print(":");
  if (minute < 10) // We need a leading "0"
  {
    Serial.print("0");
  }
  Serial.print(minute);
  Serial.print(":");
  if (second < 10) // We need a leading "0"
  {
    Serial.print("0");
  }
  Serial.print(second);
  Serial.print(" ");
  Serial.print(month);
  Serial.print("/");
  Serial.print(dayOfMonth);
  Serial.print("/");
  Serial.print(year);
  Serial.print(F(" Day of week: "));
  switch (dayOfWeek) {
    case 1:
      Serial.println(F("Sunday"));
      break;
    case 2:
      Serial.println(F("Monday"));
      break;
    case 3:
      Serial.println(F("Tuesday"));
      break;
    case 4:
      Serial.println(F("Wednesday"));
      break;
    case 5:
      Serial.println(F("Thursday"));
      break;
    case 6:
      Serial.println(F("Friday"));
      break;
    case 7:
      Serial.println(F("Saturday"));
      break;
  }
}

// Convert to binary coded decimal
uint8_t toBCD(uint8_t val)
{
  return ( (val / 10 * 16) + (val % 10) );
}
// Convert from binary coded decimal
uint8_t fromBCD(uint8_t val)
{
  return ( (val / 16 * 10) + (val % 16) );
}
